from pyramid.static import static_view
from pyramid.interfaces import (
    IRouteRequest,
    IViewClassifier,
    IView,
)
from pyramid.config import not_
from pyramid.compat import string_types
from zope.interface import Interface


ANY_KEY = '*'
UNKNOWN_KEY = '<unknown>'


def _get_pattern(route):
    pattern = route.pattern

    if not pattern.startswith('/'):
        pattern = '/%s' % pattern
    return pattern


def _get_request_methods(route_request_methods, view_request_methods):
    excludes = set()

    if route_request_methods:
        route_request_methods = set(route_request_methods)

    if view_request_methods:
        view_request_methods = set(view_request_methods)

        for method in view_request_methods.copy():
            if method.startswith('!'):
                view_request_methods.remove(method)
                excludes.add(method[1:])

    has_route_methods = route_request_methods is not None
    has_view_methods = len(view_request_methods) > 0
    has_methods = has_route_methods or has_view_methods

    if has_route_methods is False and has_view_methods is False:
        request_methods = [ANY_KEY]
    elif has_route_methods is False and has_view_methods is True:
        request_methods = view_request_methods
    elif has_route_methods is True and has_view_methods is False:
        request_methods = route_request_methods
    else:
        request_methods = route_request_methods.intersection(
            view_request_methods
        )

    request_methods = set(request_methods).difference(excludes)

    if has_methods and not request_methods:
        request_methods = '<route mismatch>'
    elif request_methods:
        if excludes and request_methods == set([ANY_KEY]):
            for exclude in excludes:
                request_methods.add('!%s' % exclude)

        request_methods = ','.join(sorted(request_methods))

    return request_methods


def _get_view_module(view_callable):
    if view_callable is None:
        return UNKNOWN_KEY, ''

    if hasattr(view_callable, '__name__'):
        if hasattr(view_callable, '__original_view__'):
            original_view = view_callable.__original_view__
        else:
            original_view = None

        if isinstance(original_view, static_view):
            raise Exception()

            # skip static views
            if original_view.package_name is not None:
                return '%s:%s' % (
                    original_view.package_name,
                    original_view.docroot
                ), ''
            else:
                return original_view.docroot
        else:
            view_name = view_callable.__name__
    else:
        # Currently only MultiView hits this,
        # we could just not run _get_view_module
        # for them and remove this logic
        view_name = str(view_callable)

    view_module = '%s.%s' % (
        view_callable.__module__,
        view_name,
    )

    # If pyramid wraps something in wsgiapp or wsgiapp2 decorators
    # that is currently returned as pyramid.router.decorator, lets
    # hack a nice name in:
    if view_module == 'pyramid.router.decorator':
        view_module = '<wsgiapp>'

    return view_module, view_callable.__doc__


def get_route_data(route, registry):
    pattern = _get_pattern(route)

    request_iface = registry.queryUtility(
        IRouteRequest,
        name=route.name
    )

    route_request_methods = None
    view_request_methods_order = []
    view_request_methods = {}
    view_callable = None

    route_intr = registry.introspector.get(
        'routes', route.name
    )

    if request_iface is None:
        return [
            (route.name, _get_pattern(route), UNKNOWN_KEY, ANY_KEY)
        ]

    view_callable = registry.adapters.lookup(
        (IViewClassifier, request_iface, Interface),
        IView,
        name='',
        default=None
    )

    try:
        view_module, view_docs = _get_view_module(view_callable)
    except:
        return []

    # Introspectables can be turned off, so there could be a chance
    # that we have no `route_intr` but we do have a route + callable
    if route_intr is None:
        view_request_methods[view_module] = []
        view_request_methods_order.append(view_module)
    else:
        if route_intr.get('static', False) is True:
            return [
                (route.name, route_intr['external_url'], UNKNOWN_KEY, ANY_KEY)
            ]

        route_request_methods = route_intr['request_methods']
        view_intr = registry.introspector.related(route_intr)

        if view_intr:
            for view in view_intr:
                request_method = view.get('request_methods')

                if view.get('attr') is not None:
                    view_callable = getattr(view['callable'], view['attr'])
                    view_module, view_docs = _get_view_module(view_callable)

                if request_method is not None:
                    view_callable = view['callable']

                    view_module, view_docs = _get_view_module(view_callable)

                    if view_module not in view_request_methods:
                        view_request_methods[view_module] = []
                        view_request_methods_order.append(view_module)

                    if isinstance(request_method, string_types):
                        request_method = (request_method,)
                    elif isinstance(request_method, not_):
                        request_method = ('!%s' % request_method.value,)

                    view_request_methods[view_module].extend(request_method)
                else:
                    if view_module not in view_request_methods:
                        view_request_methods[view_module] = []
                        view_request_methods_order.append(view_module)

        else:
            view_request_methods[view_module] = []
            view_request_methods_order.append(view_module)

    final_routes = []

    for view_module in view_request_methods_order:
        methods = view_request_methods[view_module]
        request_methods = _get_request_methods(
            route_request_methods,
            methods
        )

        final_routes.append((
            route.name,
            pattern,
            view_module,
            request_methods,
            view_docs,
        ))

    return final_routes



"""
Sphinx extension that is able to convert pyramid routes to rst
"""
import sys

import docutils
from docutils import nodes
from docutils.parsers.rst import Directive, directives
from pyramid.paster import bootstrap
from pyramid.compat import PY3
from pyramid.config import Configurator
from pyramid_autodoc.utils import get_route_data, ANY_KEY
from sphinxcontrib.autohttp.common import http_directive
from sphinxcontrib import httpdomain
from docutils.statemachine import ViewList
from sphinx.util.nodes import nested_parse_with_titles
import re


class RouteDirective(Directive):
    """ Route directive.

    Injects sections in the documentation about the routes registered in the
    given module.

    Usage, in a sphinx documentation::

        .. pyramid-autodoc:: development.ini
            :skip: ^/status
            :match: ^/v1
    """
    has_content = True
    required_arguments = 1
    option_spec = {
        'match-path': directives.unchanged,
        'match-module': directives.unchanged,
        'skip-path': directives.unchanged,
        'match-module': directives.unchanged,
        'format': directives.unchanged,
    }

    def __init__(self, *args, **kwargs):
        super(RouteDirective, self).__init__(*args, **kwargs)
        self.env = self.state.document.settings.env

    def matches_pattern(self, filters, value):
        if filters is not None:
            for path_filter in filters:
                if re.match(path_filter, value):
                    return True

        return False

    def get_routes(self, ini_file, path_blacklist=None, path_whitelist=None,
                   module_blacklist=None, module_whitelist=None):
        env = bootstrap(ini_file)
        registry = env['registry']
        config = Configurator(registry=registry)
        mapper = config.get_routes_mapper()

        try:  # only supported in pyramid 1.6
            routes = mapper.get_routes(include_static=True)
        except:
            routes = mapper.get_routes()

        mapped_routes = []

        for route in routes:
            route_data = get_route_data(route, registry)
            for name, pattern, view, method, docs in route_data:
                if path_blacklist:
                    if self.matches_pattern(path_blacklist, pattern):
                        continue

                if path_whitelist:
                    if not self.matches_pattern(path_whitelist, pattern):
                        continue

                if module_blacklist:
                    if self.matches_pattern(module_blacklist, view):
                        continue

                if module_whitelist:
                    if not self.matches_pattern(module_whitelist, view):
                        continue

                mapped_routes.append({
                    'name': name,
                    'pattern': pattern,
                    'view': view,
                    'method': method,
                    'docs': trim(docs),
                })

        return mapped_routes

    def make_httpdomain_rst(self, mapped_routes):
        node = nodes.section()
        node.document = self.state.document
        result = ViewList()

        for route in mapped_routes:
            if route['method'] == ANY_KEY:
                method = 'any'
            else:
                method = route['method']

            directives = http_directive(
                method,
                route['pattern'],
                route['docs'],
            )

            for line in directives:
                result.append(line, '<autopyramid>')

        nested_parse_with_titles(self.state, result, node)

        return node.children

    def make_custom_rst(self, mapped_routes):
        custom_nodes = []

        for mapped_route in mapped_routes:
            env = self.state.document.settings.env
            route_id = "route-%d" % env.new_serialno('route')
            route_node = nodes.section(ids=[route_id])
            title = mapped_route['pattern']

            route_node += nodes.title(text=title)

            real_table = nodes.table('')
            group = nodes.tgroup('', cols=2)
            real_table += group
            group += nodes.colspec('', colwidth=10)
            group += nodes.colspec('', colwidth=90)
            body = nodes.tbody('')
            group += body

            def get_row(*column_texts):
                row = nodes.row('')
                for text in column_texts:
                    node = nodes.paragraph('', '', nodes.Text(text))
                    row += nodes.entry('', node)

                return row

            body += get_row('Module', mapped_route['view'])
            body += get_row('Request Method', mapped_route['method'])
            body += get_row('Route Name', mapped_route['name'])

            route_node.append(real_table)

            if mapped_route['docs']:
                route_node += rst2node(
                    mapped_route['view'], mapped_route['docs']
                )

            custom_nodes.append(route_node)

        return custom_nodes

    def run(self):
        ini_file = self.arguments[0]
        fmt = self.options.get('format', 'custom')
        path_blacklist = self.options.get('skip-path', '').split() or None
        path_whitelist = self.options.get('match-path', '').split() or None
        module_blacklist = self.options.get('skip-module', '').split() or None
        module_whitelist = self.options.get('match-module', '').split() or None

        routes = self.get_routes(
            ini_file,
            path_blacklist=path_blacklist,
            path_whitelist=path_whitelist,
            module_blacklist=module_blacklist,
            module_whitelist=module_whitelist,
        )

        if fmt == 'custom':
            return self.make_custom_rst(routes)
        elif fmt == 'httpdomain':
            return self.make_httpdomain_rst(routes)
        else:
            raise Exception('Unsupported format %s' % fmt)


def trim(docstring):
    """
    Remove the tabs to spaces, and remove the extra spaces / tabs that are in
    front of the text in docstrings.

    Implementation taken from http://www.python.org/dev/peps/pep-0257/
    """
    if not docstring:
        return ''
    # Convert tabs to spaces (following the normal Python rules)
    # and split into a list of lines:
    lines = docstring.expandtabs().splitlines()
    # Determine minimum indentation (first line doesn't count):
    indent = sys.maxsize
    for line in lines[1:]:
        stripped = line.lstrip()
        if stripped:
            indent = min(indent, len(line) - len(stripped))
    # Remove indentation (first line is special):
    trimmed = [lines[0].strip()]
    if indent < sys.maxsize:
        for line in lines[1:]:
            trimmed.append(line[indent:].rstrip())
    # Strip off trailing and leading blank lines:
    while trimmed and not trimmed[-1]:
        trimmed.pop()
    while trimmed and not trimmed[0]:
        trimmed.pop(0)
    # Return a single string:
    res = '\n'.join(trimmed)
    if not PY3 and not isinstance(res, unicode):
        res = res.decode('utf8')
    return res


class Env(object):
    temp_data = {}
    docname = ''


def rst2node(doc_name, data):
    """Converts a reStructuredText into its node
    """
    if not data:
        return
    parser = docutils.parsers.rst.Parser()
    document = docutils.utils.new_document('<%s>' % doc_name)
    document.settings = docutils.frontend.OptionParser().get_default_values()
    document.settings.tab_width = 4
    document.settings.pep_references = False
    document.settings.rfc_references = False
    document.settings.env = Env()
    parser.parse(data, document)
    if len(document.children) == 1:
        return document.children[0]
    else:
        par = docutils.nodes.paragraph()
        for child in document.children:
            par += child
        return par


def setup(app):
    """Hook the directives when Sphinx ask for it."""
    if 'http' not in app.domains:
        httpdomain.setup(app)

    app.add_directive('autopyramid', RouteDirective)
